#include <stdio.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/relaxng.h>
#include <libxml/xmlschemas.h>
#include <libxml/xpath.h>

static xmlNode **xpath_nodeset(xmlXPathContext *ctx, xmlNode *parent, xmlChar *expr)
{
	xmlNode **ret;
	xmlNode *orig = NULL;
	xmlXPathObject *obj;

	if (parent) {
		orig = ctx->node;
		ctx->node = parent;
	}
	obj = xmlXPathEvalExpression(expr, ctx);
	if (orig)
		ctx->node = orig;

	if (!obj || obj->type != XPATH_NODESET || !obj->nodesetval) {
		if (obj)
			xmlXPathFreeObject(obj);
		return NULL;
	}

	ret = calloc(sizeof(xmlNode *), obj->nodesetval->nodeNr + 1);
	memcpy(ret, obj->nodesetval->nodeTab, obj->nodesetval->nodeNr * sizeof(xmlNode *));

	xmlXPathFreeObject(obj);
	return ret;
}


typedef struct {
	xmlDoc *xml_doc;

	/* RelaxNG contexts */
	xmlRelaxNGParserCtxt *rng_pctx;
	xmlRelaxNGValidCtxt *rng_vctx;
	xmlRelaxNG *rng;
	xmlDoc *rng_doc;

	/* Schema contexts */
	xmlSchemaParserCtxt *xsd_pctx;
	xmlSchemaValidCtxt *xsd_vctx;
	xmlSchema *xsd;
	xmlDoc *xsd_doc;

	xmlXPathContext *xpath_ctx;

	xmlNode *root;
} test_st;

static void init(test_st *st, char *doc, char *rng, char *xsd)
{
	memset(st, 0, sizeof(*st));

	/* create RelaxNG validation contexts */
	xmlRelaxNGInitTypes();
	if (!(st->rng_doc = xmlParseFile(rng)) ||
	    !(st->rng_pctx = xmlRelaxNGNewDocParserCtxt(st->rng_doc)) ||
	    !(st->rng = xmlRelaxNGParse(st->rng_pctx)) ||
	    !(st->rng_vctx = xmlRelaxNGNewValidCtxt(st->rng))) {
		fprintf(stderr, "cannot open RelaxNG schema file \"%s\"\n", rng);
	}

	/* create Schema validation contexts */
	if (!(st->xsd_doc = xmlParseFile(xsd)) ||
	    !(st->xsd_pctx = xmlSchemaNewDocParserCtxt(st->xsd_doc)) ||
	    !(st->xsd = xmlSchemaParse(st->xsd_pctx)) ||
	    !(st->xsd_vctx = xmlSchemaNewValidCtxt(st->xsd))) {
		fprintf(stderr, "cannot open XSD schema file \"%s\"\n", xsd);
	}

	/* parse document */
	if (!(st->xml_doc = xmlParseFile(doc))) {
		fprintf(stderr, "cannot open XML file \"%s\"\n", doc);
		exit(-1);
	}
	
	if (!(st->xpath_ctx = xmlXPathNewContext(st->xml_doc))) {
		fprintf(stderr, "unable to create XPath context for file \"%s\"\n", doc);
		exit(-1);
	}

	/* get /test */
	if (!(st->root = xmlDocGetRootElement(st->xml_doc))) {
		fprintf(stderr, "can't find root element in \"%s\"\n", doc);
		exit(-1);
	}
}

static void valid(test_st *st, char *pfx)
{
	if (st->rng_vctx)
		fprintf(stderr, "%s xmlRelaxNGValidateDoc -> %d\n",
			pfx, xmlRelaxNGValidateDoc(st->rng_vctx, st->xml_doc));
	
	if (st->xsd_vctx)
		fprintf(stderr, "%s xmlSchemaValidateDoc -> %d\n",
			pfx, xmlSchemaValidateDoc(st->xsd_vctx, st->xml_doc));
}

int main(int argc, char **argv)
{
	test_st st;
	xmlNode **node, **refnode, *root, *new_ref = NULL, *new_node;
	xmlAttr **idref_attr;
	xmlBuffer *nodebuf, *refbuf;

	if (argc < 4) {
		fprintf(stderr, "Usage: %s <XML file> <RelaxNG schema> <XSD schema>\n", argv[0]);
		exit(-1);
	}

	init(&st, argv[1], argv[2], argv[3]);

	valid(&st, "[initial]");

	/* get /test/node */
	node = xpath_nodeset(st.xpath_ctx, NULL, (xmlChar *)"id('test-id')");
	if (!node || !*node) {
		fprintf(stderr, "can't find node \"/test/refnode[@id = 'test-id']\"\n");
		exit(-1);
	}

	/* save <node> in buffer */
	nodebuf = xmlBufferCreate();
	if (xmlNodeDump(nodebuf, st.xml_doc, *node, 0, 0) < 0) {
		fprintf(stderr, "couldn't dump node to buffer\n");
		exit(-1);
	}

	/* get /test/refnode */
	refnode = xpath_nodeset(st.xpath_ctx, NULL, (xmlChar *)"/test/refnode[@id = 'test-id']");
	if (!refnode || !*refnode) {
		fprintf(stderr, "can't find node \"/test/refnode[@id = 'test-id']\"\n");
		exit(-1);
	}

	/* save <refnode> in buffer */
	refbuf = xmlBufferCreate();
	if (xmlNodeDump(refbuf, st.xml_doc, *refnode, 0, 0) < 0) {
		fprintf(stderr, "couldn't dump node to buffer\n");
		exit(-1);
	}

	/* get <refnode>'s idref attribute */
	idref_attr = (void *)xpath_nodeset(st.xpath_ctx, *refnode, (xmlChar *)"@id");
	if (!idref_attr || !*idref_attr) {
		fprintf(stderr, "can't find IDREF to \"id('test-id')\"\n");
		exit(-1);
	}

	/* remove <refnode>'s idref */
	xmlRemoveRef(st.xml_doc, *idref_attr);

	/* remove <refnode> */
	xmlUnlinkNode(*refnode);
	xmlFreeNode(*refnode);

	/* remove <node> */
	xmlUnlinkNode(*node);
	xmlFreeNode(*node);

	/* re-validate document */
	st.root->parent->psvi = NULL;
	valid(&st, "[remove]");

	/* re-insert <node> */
	if ((xmlParseInNodeContext(st.root, (char *)nodebuf->content, nodebuf->use,
	     			   0, &new_node) != XML_ERR_OK) 
	    || !new_node
	    || !xmlAddChild(st.root, new_node)) {
		fprintf(stderr, "couldn't re-insert new_node\n");
		exit(-1);
	}

	/* re-validate document */
	new_node->parent->psvi = NULL;
	valid(&st, "[insert new_node]");

	/* re-insert <refnode> */
	if ((xmlParseInNodeContext(st.root, (char *)refbuf->content, refbuf->use,
	     			   0, &new_ref) != XML_ERR_OK)
	    || !new_ref
	    || !xmlAddNextSibling(new_node, new_ref)) {
		fprintf(stderr, "couldn't re-insert new_ref\n");
		exit(-1);
	}

	/* re-validate document */
	new_ref->parent->psvi = NULL;
	valid(&st, "[insert new_ref]");

	printf("last error: %p\n", xmlGetLastError());
	return 0;
}

